/*
 *  @version 	        @(#)bongo.java	1.9 98/05/15
 *
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package locators;
import suntest.javastar.lib.*;
import java.awt.AWTEvent;
import java.awt.event.*;
import java.awt.Component;
import java.awt.Point;
import java.util.Vector;
import java.util.Hashtable;
import marimba.gui.*;
import locators.utils.BongoInfo;

/**
* <ul>
* <li>Non Component Locator for Bongo toolkit
* <li>Copyright (C) 1997 Sun Microsystem Inc.
* <li>Bongo and Marimba are registered trademarks of Marimba, Inc
* </ul>
* <p> The bongo class implements a nonComponent Locators solution for
* Marimba's Bongo toolkit.  When used in conjunction with JavaStar the
* locator allows JavaStar tests to operate in the bongo object level.
* This version of the locators support index, captions, internal names, and
* non hiearchical recording.
* Index recording creates the tags by determining the index of a component
* within its class.  For examples a TextField is the 2nd textfield in the
* Container based upon its creation order.
* The internal name in bongo is used when the code does a .setName on
* a bongo widget.  This is generally the most reliable method of identifying
* an object.  If an internal name is not unique the locator does not use
* it for identification.
* A Caption of a bongo component is used when an object is of a certain
* class that contains a caption.  If the caption is not unique within the
* container instead of index JavaStar will use the caption.  Non hiearchical
* recording occurs when a component has a unique caption within the entire
* application.  In this case the hiearchy does not have to be recorded.
* When recording a test internal names have the highest precedence, then
* captions are used, followed by index.  If a component has a unique caption
* throughout the application, the non-hiearchical tag is used.  When coding
* a script manually any of these approaches may be utilized.
*
* @author     Dan Schaffer
* @version    @(#)bongo.java	1.9 0
*/
public class bongo implements JSNonComponentLocator {
  /** character for indexes: Widget%1 */
  public final static String INDEXSEP="%";
  /** character for captions: ButtonWidget^ok^ */
  public final static String CAPTIONSEP="^";
  /** character for internal names */
  public final static String INTERNALSEP="@";

  /**
   * Returns the non-component object and a string representation that
   * is recreating by getNamedObjectData.  If a unique .getName exists
   * for the bongo object that is returned otherwise creates name by
   * moving down widget tree using indexes.  For examples:
   * "PopupWidget.Presentation%0.FolderWidget%0.PageWidget%1.CheckBoxWidget%1"
   *
   * @param    comp The AWT based Component or frame of Bongo, its a PlayerPanel
   * @param    e The AWTEvent, extract x,y coords
   * @return   JSNCLData object is (String name,Point(widgets loc),Object)
   * @see bongo#getNamedObjectData
   */
  public JSNCLData findObject(Component comp, AWTEvent evt){
    String name=null;
    int x,y;
    PlayerPanel pp;
    Widget w;
    BongoInfo wi=null;
    Vector vTree;
    Point p=null;
    PopupMenu popup;

    try {
    if((comp instanceof PlayerPanel)==false)
      return null;
    // get the x,y location of the event inside the PlayerPanel
    // locateWidget returns the bongo Widget x,y
    x = ((MouseEvent)evt).getX();
    y = ((MouseEvent)evt).getY();
    pp=(PlayerPanel)comp;
    w=pp.locateWidget(x,y);
    // Attempt to get the internal name
    // Otherwise create name by walking the tree, failure in all cases returns
    // null and the frame based coordinates are used
    if (w instanceof PopupMenuItemWidget) {
      popup=((PopupMenuItemWidget)w).getMenu();
      if (popup!=null && popup.owner!=null)
	w=popup.owner;
    }
    w=walkUpUntilEnabled(w);
    vTree=buildTree(pp);
    wi=findWidget(vTree,w);
    if (wi==null)
      return null;
    name=wi.theTag;
    p=getLocation(w);
    } catch (Exception e){
      e.printStackTrace();
    }
    return new JSNCLData(name,p,wi.widget);
  }

  /**
   * Returns the non-component object after getting the actual Widget from
   * the string description passed in.  Reconstructs the widget from the
   * text parameter.  Looks to see if its a internal name otherwise attempts
   * to find object in widget tree.
   *
   * @param    c The AWT based Component or frame of Bongo, its a PlayerPanel
   * @param    name The name of the object
   * @return   JSNCLData object is (String name,Point(widgets loc),Object)
   * @see bongo#findObject
   */
  public JSNCLData getNamedObjectData(Component c, String name){
    // If the component is not PlayerPanel fail
    // its not a bongo widget
    if((c instanceof PlayerPanel)==false)
      return null;
    PlayerPanel pp=(PlayerPanel)c;

    // creates widget tree, then looks for name as internal otherwise looks for
    // name as tag
    Widget w;
    Vector vTree=null;
    BongoInfo wi=null;
    Point p=null;
    try {
    vTree=buildTree(pp);
    if (wi==null) {
      wi=findTag(vTree,name);
    }
    if (wi==null) {
      return null;
    }
    if (isVisible(wi.widget)==false || isEnabled(wi.widget)==false)
      return null;
    p=getLocation(wi.widget);
    } catch (Exception e) { e.printStackTrace(); }
    return new JSNCLData(name,p,wi.widget);
  }

  /**
   * walks up widget tree until find a widget that is enabled. If
   * no widgets are enabled even at top level null is returned.
   * @param v the widget to start at
   * @returns the first widget that is enabled or null
   */
  private static Widget walkUpUntilEnabled(Widget v) {
    boolean enabled=false;
    while(enabled==false) {
      enabled=isEnabled(v);
      if (enabled==false) {
	v=v.parent;
	if (v==null)
	  return null;
      }
    }
    return v;
  }
  /** returns whether the widget is visible.
   * @param v the widget to test
   * @returns boolean true if visible, false if not visible
   */
  private static boolean isVisible(Widget v) {
    return v.isVisible();
  }
  /** returns whether the widget is enabled
   * @param v the widget to test
   * @returns true if widget is enabled, false if widget is disabled
   */
  private static boolean isEnabled(Widget v) {
    return !v.isDisabled();
  }
  /** returns the position of the widget relative to the PlayerFrame
   * walks up the widget tree adding the relative positions of each widget
   * @param w the widget
   * @returns Point position relative to the PlayerPanel
  */
  public static Point getLocation(Widget w) {
    int x,y;
    Point p=new Point(0,0);
    while (w!=null) {
      x=w.x;
      y=w.y;
      if (x<0) x=0;
      if (y<0) y=0;
      p=new Point(p.x+w.x,p.y+w.y);
      w=w.parent;
    }
    return p;
  }
  /** Searches the widget for a child widget matching the class.  Returns
  * the widget when find the child widget with the correct index
  * @param w the parent widget
  * @param cname the string class of the child
  * @param index the number child of the particular class to search for
  * @returns Widget the child widget matching index and class or null if
  * no child matches
  */
  public static Widget findIndexWidget(Widget w,String cname,int index) {
    Widget wRet=null;
    int iNum=0;
    for (int i=0;i<w.nwidgets;i++){
      if (getWidgetClassName(w.widgets[i]).equals(cname))
	iNum++;
      if (iNum>index)
	return w.widgets[i];
    }
    return wRet;
  }
  /** Creates a string representation of the widget.  For each level in
  *   the widget hiearchy get the index of the component and the class.
  *   For each level create <classname>%<index>.  The "." separates each
  *   widget.
  * @param w the widget the create the tag
  * @returns String the tag built using hierarchy and indexes
  */
  public static String buildTag(Widget w) {
    int index;
    String sRet="",s;
    // starting at top get the widget class,index append to tag and walk
    // down to top level where w.parent is null
    // if can't find a component is the parent list return null
    while (w!=null) {
      s=getWidgetClassName(w);
      if (w.parent!=null) {
	index=getWidgetIndex(w);
	if (index!= -1)
	  s+=INDEXSEP+index;
      }
      if (sRet=="")
	sRet=s;
      else
	sRet=s+"."+sRet;
      w=w.parent;
    }
    return sRet;
  }
  /** returns the index of this widget from its parents widget list by class name.  The parent of the widget calculates the index by classname of the widget.
   * @param w the child widget
   * @returns int the index
   */
  public static int getWidgetIndex(Widget w) {
    Widget container=w.parent;
    int index=0;
    if (container==null)
      return -1;
    for (int i=0;i<container.nwidgets;i++) {
      if (container.widgets[i]==w)
	return index;
      if (container.widgets[i].getClass()==w.getClass())
	index++;
    }
    return -1;
  }
  /** returns the caption of the widget.  Only ButtonWidget and subclasses
   * will look for captions.  Prior this method attempting to use reflection
   * to look for .getLabel available method but this proved to be too much
   * of a performance hit.  If not a subclass of ButtonWidget null is returned.
   * @param w the widget the look for caption
   * @returns String the caption or null is no caption is available
   */
  public static String getWidgetCaption(Widget w) {
    if (w instanceof ButtonWidget) {
      return ((ButtonWidget)w).getLabel();
    }
    return null;
  }
  /** gets the class name from a widget. strips marimba.gui off front and
   * if an "_" is found at end strip this off as well.
   * @param Object the class
   * @returns String the class name
   */
  public static String getWidgetClassName(Object o){
    String cl = o.getClass().getName();
    cl = cl.substring(1+cl.lastIndexOf('.'));
    // hackaround for Bongo returning script objects
    if (cl.indexOf("_") != -1) {
      cl=cl.substring(cl.indexOf("_")+1,cl.lastIndexOf("_"));
    }
    return cl;
  }
  /** walks up hiearchy tree to get top level widget. The top level
   * widget is below the PlayerPanel.
   * @param w the widget to start
   * @returns the top level widget
   */
  public static Widget getRoot(Widget w) {
    Widget wlast=w;
    if (w==null)
      return null;
    while (w!=null) {
      wlast=w;
      w=w.parent;
    }
    return wlast;
  }
  /** returns the BongoInfo in the Vector widgetTree with the
   *  matching tag name
   * @param v the entire widget tree
   * @param the tag to search for
   * @returns BongoInfo the data structure containing the matching
   * widget
   */
  public static BongoInfo findTag(Vector v,String s) {
    BongoInfo wi;
    for (int i=0;i<v.size();i++) {
      wi=(BongoInfo)v.elementAt(i);
      if ((wi.intName!=null && wi.intName.equals(s)) ||
	  (wi.parent+"."+wi.className+INDEXSEP+wi.index).equals(s) ||
	  (wi.caption!=null && (wi.parent+"."+wi.className+CAPTIONSEP+wi.caption+CAPTIONSEP).equals(s)) ||
	  (wi.localTagName!=null && wi.localTagName.equals(s))) {
	return wi;
      }
      // supports old internal names
      if (removeMarkers(wi.theTag,INTERNALSEP).equals(s)) {
	return wi;
      }
    }
    return null;
  }
  /** return the BongoInfo in the Vector widgetTree with the
  *  matching widget
  * @param v the entire widget tree
  * @param w the widget to find
  * @returns BongoInfo the data structure containing the matching
  * widget with its tag information
  */
  public static BongoInfo findWidget(Vector v,Widget w) {
    BongoInfo wi;
    for (int i=0;i<v.size();i++) {
      wi=(BongoInfo)v.elementAt(i);
      if (wi.widget==w)
	return wi;
    }
    return null;
  }
  /** walks through component tree building list of all
  * current bongo widgets
  * @param pp the root Component containing bongo objects
  * @returns Vector the tree created
  */
  public static Vector buildTree(PlayerPanel pp) {
    int index,i;
    Hashtable vIntName=new Hashtable();
    Hashtable vCaption=new Hashtable();
    PopupMenu popup;
    Vector vRet=new Vector();
    Vector vCon=new Vector();
    BongoInfo w,wi;
    Widget wRoot=getRoot(pp.locateWidget(0,0));
    vCon.addElement(new BongoInfo(wRoot,null));
    while (vCon.isEmpty()==false) {
      w=(BongoInfo)vCon.firstElement();
      vCon.removeElementAt(0);

      w.intName=w.widget.getName();
      if (w.intName!=null)
	w.intName=INTERNALSEP+w.intName+INTERNALSEP;
      w.className=getWidgetClassName(w.widget);
      w.index=getWidgetIndex(w.widget);
      w.caption=getWidgetCaption(w.widget);
      if (w.caption!=null) {
	w.localTagName=w.className+CAPTIONSEP+w.caption+CAPTIONSEP;
	w.fullTagName=w.localTagName;
	if (w.parent!=null) {
	  w.fullTagName=w.parent+"."+w.fullTagName;
	}
      }
      else {
	w.localTagName=null;
	if (w.index != -1) {
	  w.fullTagName=w.className+INDEXSEP+w.index;
	} else {
	  w.fullTagName=w.className;
	}
	if (w.parent!=null) {
	  w.fullTagName=w.parent+"."+w.fullTagName;
	}
      }
      if (w.caption!=null) {
	if (vCaption.get(w.localTagName)!=null) {
	  i=((Integer)vCaption.get(w.localTagName)).intValue();
	  wi=(BongoInfo)vRet.elementAt(i);
	  wi.localTagName=null;
	  wi.setTheTag();
	  vRet.setElementAt(wi,i);
	  w.localTagName=null;
	} else {
	  vCaption.put(w.localTagName,new Integer(vRet.size()));
	  vCaption.put(w.fullTagName,new Integer(vRet.size()));
	}
	if (w.localTagName==null) {
	  if (vCaption.get(w.fullTagName)!=null) {
	    i=((Integer)vCaption.get(w.fullTagName)).intValue();
	    wi=(BongoInfo)vRet.elementAt(i);
	    wi.fullTagName=wi.className+INDEXSEP+wi.index;
	    if (wi.parent!=null) {
	      wi.fullTagName=wi.parent+"."+wi.fullTagName;
	    }
	    wi.setTheTag();
	    vRet.setElementAt(wi,i);
	    w.fullTagName=w.className+INDEXSEP+w.index;
	    if (w.parent!=null) {
	      w.fullTagName=w.parent+"."+w.fullTagName;
	    }
	  }
	}
      }

      if (w.intName!=null) {
	if (vIntName.get(w.intName)!=null) {
	  i=((Integer)vIntName.get(w.intName)).intValue();
	  wi=(BongoInfo)vRet.elementAt(i);
	  wi.intName=null;
	  wi.setTheTag();
	  vRet.setElementAt(wi,i);
	  w.intName=null;
	} else {
	  vIntName.put(w.intName,new Integer(vRet.size()));
	}
      }
      w.setTheTag();
      vRet.addElement(w);
      for (i=0;i<w.widget.nwidgets;i++)
	vCon.addElement(new BongoInfo(w.widget.widgets[i],w.fullTagName));
      if (w.widget instanceof MenuButtonWidget) {
	popup=((MenuButtonWidget)w.widget).getMenu();
	for (i=0;i<popup.getItemCount();i++) {
	  vCon.addElement(new BongoInfo(popup.getItem(i),w.fullTagName));
	}
      }
    }
    return vRet;
  }
  /** removes markers from the tag such as internal names, captions.  Also
   * if the string containing the special character, two characters are
   * duplicated, escaping the character.  The two characters are changed
   * back to one character.  The marker is removed from the front and
   * end of the tag.
   * @param tag the tag
   * @param marker the marker symbol
   * @returns the tag without the marker
   */
  public static String removeMarkers(String tag,String marker) {
    if (tag.startsWith(marker))
      tag=tag.substring(1);
    if (tag.endsWith(marker))
      tag=tag.substring(0,tag.length()-1);
    tag=replaceString(tag,marker+marker,marker);
    return tag;
  }
  /** markers are added to the front and back of the tag.  If the marker
   * character is found in the tag it is duplicated, escaping the character.
   * @param the tag
   * @param the marker
   * @returns the tag with markers added
   */
  public static String addMarkers(String tag,String marker) {
    tag=replaceString(tag,marker,marker+marker);
    tag=marker+tag+marker;
    return tag;
  }
  /** replaces a string with another string for all occurrences in a string.
   * @param s the source string
   * @param replace what to replace
   * @param replaceWith what to replace the string with
   */
  public static String replaceString(String s,String replace,String replaceWith) {
    int index=0;
    for (index=s.indexOf(replace,index);index!= -1;) {
      s=s.substring(0,index)+replaceWith+s.substring(index+replace.length());
      index=index-replace.length()+replaceWith.length()+1;
      index=s.indexOf(replace,index);
    }
    return s;
  }
  /** Used for Debugging purpose.  Outputs to stdout the entire tree.
   * this is really going to hurt performance but is useful for debugging
   * purpose.
   * @param pp the player panel Component.  Parent to all bongo Frames
   */
  public static void printTree(PlayerPanel pp) {
    Vector vWid=buildTree(pp);
    printTree(vWid);
  }
  /** Used for Debugging purpose.  Outputs to stdout the entire tree.
   * this is really going to hurt performance but is useful for debugging
   * purpose.
   * @param vWid the widget tree to print out
   */
  public static void printTree(Vector vWid) {
    BongoInfo wi;
    for (int i=0;i<vWid.size();i++) {
      wi=(BongoInfo)vWid.elementAt(i);
      System.out.println(wi);
    }
  }
}
